#ifndef ASTL_CURSORS
#define ASTL_CURSORS

#include <stack>
#include <list>
#include <hash_set>

template <class DFA>
class cursor
{
public:
  typedef cursor              self;
  typedef typename DFA::State State;
  typedef typename DFA::Tag   Tag;

protected:
  const DFA *dfa;
  State q;

public:
  cursor()
    : dfa(NULL)
  { }

  cursor(const DFA &a)
    : dfa(&a), q(dfa->initial())
  { }

  cursor(const DFA &a, State p)
    : dfa(&a), q(p)
  { }

  State src() const {
    return q;
  }

  self& operator= (State p) {
    q = p;
    return *this;
  }

  self& operator= (const self &x) {
    q = x.q;
    dfa = x.dfa;
    return *this;
  }

  bool operator== (const self &x) const {
    return q == x.q;
  }

  bool sink() const {
    return q == dfa->null_state;
  }

  bool forward(unsigned int letter) {
    q = dfa->delta1(q, letter);
    return q != dfa->null_state;
  }

  bool src_final() const {
    return dfa->final(q);
  }

  const Tag& src_tag() const  {
    return dfa->tag(q);
  }
};

template <class DFA>
class forward_cursor : public cursor<DFA>
{
public:
  typedef forward_cursor      self;
  typedef cursor<DFA>         super;
  // Types State and Tag are inherited from cursor class

protected:
  typedef typename DFA::Edges::const_iterator edges_iterator;
  edges_iterator transition;

public:
  forward_cursor(const DFA &a, State p, unsigned int Letter) 
    : super(a, p), transition(dfa->delta2(p).find(Letter)) 
  { }
  
  forward_cursor(const DFA &a, State p, edges_iterator t)
    : super(a, p), transition(t)
  { }
  
  forward_cursor(const DFA &a, State p)
    : super(a, p)
  { }

  forward_cursor(const DFA &a)
    : super(a)
  { }

  forward_cursor()
    : super()
  { }

  self& operator= (State p) {
    q = p;
    return *this;
  }

  self& operator= (const self &x) {
    dfa = x.dfa;
    q = x.q;
    transition = x.transition;
    return *this;
  }

  bool operator== (const self &x) const {
    return (q == x.q && transition == x.transition);
  }

  unsigned int letter() const {
    return (*transition).first;
  }

  bool sink() const {
    return q == dfa->null_state || transition == dfa->delta2(q).end();
  }

  bool first_transition() {
    return (transition = dfa->delta2(q).begin()) != dfa->delta2(q).end();
  }
  
  bool next_transition() {
    return ++transition != dfa->delta2(q).end();
  }

  bool forward(unsigned int letter) {
    return super::forward(letter);
  }

  void forward() {
    q = (*transition).second;
  }

  bool find(unsigned int letter)  {
    if (dfa->delta1(q, letter) != dfa->null_state) {
      transition = dfa->delta2(q).find(letter);
      return true;
    }
    return false;
  }

  State aim() const {
    return (*transition).second;
  }

  bool aim_final() const {
    return dfa->final((*transition).second);
  }

  const Tag& aim_tag() const {
    return dfa->tag((*transition).second);
  }

  // Not a standard requirement:
  friend ostream& operator << (ostream &out, const self &x) {
    if (x.transition != x.dfa->delta2(x.q).end())
      out << x.q << " '" << (*(x.transition)).first << "' " << (*(x.transition)).second;
    else
      out << x.q << " 'null' 0";
    return out;
  }

  // Not a standard requirement:
  bool operator < (const self &x) const {   
    // STL containers need operator < on
    // their value_type (to define < on themselves)
    return q < x.q || (q == x.q && transition < x.transition);
  }
};

template <class ForwardCursor>
class stack_cursor : protected stack<ForwardCursor, list<ForwardCursor> >
{
public:
  typedef stack_cursor self;
  typedef stack<ForwardCursor, list<ForwardCursor> > super;
  //  typedef typename ForwardCursor::Tag Tag;
  typedef typename ForwardCursor::State State;

  stack_cursor(const ForwardCursor &x)
    : super() {
    super::push(x);
  }
  
  stack_cursor()  // Empty stack
    : super()
  { }

  State src() const {
    return top().src();
  }

  State aim() const {
    return top().aim();
  }

  unsigned int letter() const {
    return top().letter();
  }

  bool sink() const {
    return top.sink();
  }

  bool find(unsigned int a) {
    return top().find(a);
  } 

  bool first_transition() {
    return top().first_transition();
  }

  bool next_transition() {
    return top().next_transition();
  }

  // push:
  void forward() {
    super::push(top());
    top().forward();
  }

  // push:
  bool forward(unsigned int a) {
    if (top().find(a)) {
      super::push(top());
      top().forward();
      return true;
    }
    return false;
  }
 
  // pop:
  // return false if RESULTING stack is empty
  bool backward() {
    super::pop();
    return !empty();
  }

  bool aim_final() const {
    return top().aim_final();
  }

  bool src_final() const {
    return top().src_final();
  }

  // Compare entire stacks
  bool operator == (const self &x) const {
    return (super) *this == (super) x;
  }

  /*
  const Tag& src_tag() const {
    return top().src_tag();
  }

  const Tag& aim_tag() const {
    return top().aim_tag();
  }
  */
};
  
// Depth first cursor base
template <class StackCursor, class MarkerFunction>
class dfirst_cursor_base
{
protected:
  StackCursor c;
  bool has_poped;
  MarkerFunction visited;

public:
  typedef dfirst_cursor_base          self;
  //  typedef typename StackCursor::Tag   Tag;
  typedef typename StackCursor::State State;

  dfirst_cursor_base(const StackCursor &x)
    : c(x), has_poped(false), visited()
  {  
    visited(c);   // set c to 'visited'
  }

  dfirst_cursor_base()
    : c(), has_poped(false)
  { }

  // Move forward on current transition
  // Position on first_transition() of aim
  // Return true if push, false otherwise
  bool forward() {
    if (has_poped) {
      // move on to the next outgoing transition if any otherwise pop
      if (!c.next_transition()) {
	// at edges end
	return !c.backward();
      }
      else {
	has_poped = false;
	return true;
      }
    }
    c.forward();
    if (visited(c) || !c.first_transition()) {
      // cannot forward so backward => pop
      c.backward();
      has_poped = true;
      return false;
    }
    return true;
  }

  unsigned int letter() const {
    return c.letter();
  }

  bool aim_final() const {
    return c.aim_final();
  }

  bool src_final() const {
    return c.src_final();
  }
  
  State src() const {
    return c.src();
  }
  
  State aim() const {
    return c.aim();
  }

  // Semantic: a dfirst_cursor represents an algorithm stage
  // hence the operator== behavior : it compares entire stacks (see stack cursor)
  bool operator == (const self &x) const {
    return c == x.c; // && has_poped == x.has_poped;
  }

  /*
  const Tag& src_tag() const {
    return c.src_tag();
  }

  const Tag& aim_tag() const {
    return c.aim_tag();
  }
  */
};

// Marker function for tree automata
// Return false (source state has not been visited yet)
template <class StackCursor>
struct always_false : public unary_function<StackCursor, bool>
{
  bool operator() (const StackCursor &c) const {
    return false;
  }
};

// Default marker function using a hash_set<State>
// Check if the source state is in the pool of the visited states
// if not add it to the pool and return false else return true
// the behavior of the operator() is to switch the state and to return
// the state value BEFORE the switch
template <class StackCursor>
struct default_marker : unary_function<StackCursor, bool>
{
  hash_set<typename StackCursor::State, identity<typename StackCursor::State> > pool;
  bool operator() (const StackCursor &c) {
    return !pool.insert(c.src()).second;  // insert returns pair<iterator,bool>
  }
};

// Use the tag to mark the source state
// Tag must define a method bool visited() behaving like default_marker
template <class StackCursor>
struct tag_marker : unary_function<StackCursor, bool>
{
  bool operator() (StackCursor &c) {
    return c.src_tag().visited();
  }
};

// dfirst_cursor for tree like automata (no state mark)
template <class StackCursor>
class dfirst_cursor 
  : public dfirst_cursor_base<StackCursor, always_false<StackCursor> >
{
public:
  typedef dfirst_cursor self;
  typedef dfirst_cursor_base<StackCursor, always_false<StackCursor> > super;

  dfirst_cursor(const StackCursor &c)
    : super(c)
  { }

  dfirst_cursor()
    : super()
  { }
};

// dfirst_cursor_mark makes only two passes on each transition: once in
// descending stage and once in ascending stage.
// By default, the states are marked using a hashed set of states (default_marker)
// Tags can be used too (tag_marker, more efficient because of constant time access)
template <class StackCursor, class MarkerFunction = default_marker<StackCursor> >
class dfirst_cursor_mark 
  : public dfirst_cursor_base<StackCursor, MarkerFunction>
{
public:
  typedef dfirst_cursor_mark self;
  typedef dfirst_cursor_base<StackCursor, MarkerFunction> super;

  dfirst_cursor_mark(const StackCursor &c)
    : super(c)
  { }

  dfirst_cursor_mark()
    : super()
  { }
};

template <class T, class U>
ostream& operator << (ostream &out, stack<T, U> &x) {
  stack<T, U> tmp;
  for(; !x.empty(); )
    {
      tmp.push(x.top());
      out << endl << x.top();
      x.pop();
    }
  for(; !tmp.empty(); )
    {
      x.push(tmp.top());
      tmp.pop();
    }
  return out;
}

#ifdef ZORGLUB

// Backward compatibility cursor
// DO NOT USE !!!
template <class DFA>
class D_cursor
{
private:
  typedef typename DFA::State State;
  DFA &dfa;
  State current_state;

public:
  D_cursor(const DFA &a) 
    : dfa((DFA &) a), current_state(a.null_state)
  { }

  void reset() {
    current_state = dfa.initial();
  }

  bool final() const {
    return dfa.final(current_state);
  }

  friend bool operator == (const D_cursor &x, const D_cursor &y) {
    return (x.current_state == y.current_state && &(x.dfa) == &(y.dfa));
  }

  bool operator == (const State &s) const {
    return (current_state == s);
  }

  // cast operator to State
  operator State() const {
    return (current_state);
  }

  // Set the cursor on a specific state
  D_cursor& operator = (State s)
  {
    current_state = s;
    return (*this);
  }

  // Return the current state the cursor is on
  const State& operator * () const {
    return (current_state);
  }

  // Move one transition forward
  // Return true if transition with 'a' exists (if cursor has moved)
  bool forward(const typename DFA::Alphabet &a)  
  {
    State aim = dfa.delta1(current_state, a);
    if (!(aim == dfa.null_state))
    {
      current_state = aim;
      return (true);
    }
    else
      return (false);
  }

  // Move n transitions forward
  // Return true if all transitions exist 
  // (if cursor has moved over ALL the transitions)
  template <class InputIterator>
  bool forward(InputIterator first, InputIterator last)
  {
    State s;
    if (first != last)
    {
      for(s = dfa.delta1(current_state, *first++); 
          first != last && s != dfa.null_state;
          s = dfa.delta1(s, *first++));

      if (s != dfa.null_state) 
      {
        current_state = s;
        return (true);
      }
      else
        return (false);
    }
    return (true);    // ?
  }

  // Return the tag associated to the current state
  typename DFA::Tag& tag() {
    return (dfa.tag(current_state));
  }

  const typename DFA::Tag& tag() const {
    return (dfa.tag(current_state));
  }
};

#endif  // ZORGLUB

#endif  // ASTL_CURSORS

