#ifndef ASTL_CLASS_CURSORS
#define ASTL_CLASS_CURSORS

//
// Definitions:
//   Sink state:        a state handler equal to 0
//   Sink transition:   any transition leading to sink state (undefined transitions)
//
// Remarks:
// 1  Any call to a DFA method with unusable states/transitions as arguments yields
//     undefined behavior. Therefore, at any call to a cursor method other than operator==
//     the cursor must not point to a sink state/transition.
// 2  Calls to first_transition() and next_transition() returning true means the cursor is
//    on a usable transition. Consequently source, aim and letter are accessible.
//   

// Class forward_cursor: a "pointer" to a triple (q, a, p) in (Q x Aphabet x Q)
// Constructors must by default set a forward cursor on the initial
// state of DFA
template <class DFA>
class forward_cursor 
{
public:
  typedef forward_cursor self;
  typedef State;

  forward_cursor(const DFA &a);                          // points to DFA initial state
  forward_cursor(const DFA &a, State p);                 // points to state p
  forward_cursor(const DFA &a, State p, unsigned int l); // points to transition with 'l'
                                                         // user must make sure transition exist
  forward_cursor(const DFA &a, State p, DFA::Edges::const_iterator x);  // points to transition x
  void forward();                                        // forward along current transition
  bool forward(unsigned int letter);          // try to forward along the transition with 'letter'
  bool first_transition();                    // set to the first outgoing transition of current src
  bool next_transition();                     // move on to the next outgoing transition of current src
  unsigned int letter() const;                // get current transition letter a
  bool src_final() const;                     // is q a final state ?
  bool aim_final() const;                     // is p a final state ?
  bool operator== (const self &x) const;      // return (q,a,p) == (x.q,x.a,x.p)
  bool find(unsigned int letter);             // set the cursor to the
                                              // transition with 'letter' if any, otherwise stay put
  State src() const;                          // return q
  State aim() const;                          // return p
  self& operator= (State q);                  // set the cursor on the specified state

  // Optional requirements:
  typedef Tag;
  const Tag& src_tag() const;          // return q tag
  const Tag& aim_tag() const;          // return p tag
};

// A transition is a triple: (q,a,p) in Q x Sigma x Q
// An outgoing transition of state q is a couple (a,p) in Sigma x Q
// A sink transition (q,a,p) if p = sink (then a is epsilon)
// Remark: if q = sink => p = sink

// 2 types of transition: normal transition and sink transition
// each state sink transition serves as end of range for traversal of
// outgoing transitions (iteration over d2)

// Pedantic description:
// X           type of a model of forward cursor
// x, y        objects of type X
// (q,a,p)     currently pointed transition
// (q',a',p')  pointed transition after an action with side effect (non
//             const methods: forward,first and next transition, find, operator=) 
// b           a letter

// Method:        forward
// Semantic:      move along pointed transition
// Signature:     void forward();
// Expression:    x.forward();
// Return value:  none
// Precondition:  x must not point to a sink transition
// Postcondition: resulting source state is the former aim state (q'=p). The
//                newly pointed outgoing transition is undefined until
//                a successful call to either first_transition or find.
// Complexity:    constant

// Method:        forward with letter
// Semantic:      move along q's outgoing transition labelled with letter b if any
// Signature:     bool forward(unsigned int b);
// Expression:    x.forward('a');
// Return value:  true if (q,b,d1(q,b)) exist, false otherwise
// Precondition:  q != sink
// Postcondition: if the transition is defined, q'=d1(q,b), and the
//                newly pointed outgoing transition is undefined until
//                a successful call to either first_transition or find.
//                otherwise q'=q and the pointed outgoing transition
//                is undefined until a successful call to either first_transition or find.
// Complexity:    log(d2(q))

// Method:        first_transition
// Semantic:      point to first transition of set d2(q)
// Signature:     bool first_transition();
// Expression:    x.first_transition();
// Return value:  true if d2 not empty set, false otherwise
// Precondition:  q != sink
// Postcondition: if d2 not empty set, x point to q's first outgoing
//                transition (q,a',p') otherwise pointed outgoing
//                transition is undefined.
// Complexity:    constant



// Remark: d2(q) end of range is a sink transition

// Method:        next_transition
// Semantic:      point to next transition of set d2(q)
// Signature:     bool next_transition();
// Expression:    x.next_transition();
// Return value:  true if not at end of range, false otherwise
// Precondition:  x must not point to a sink transition
// Postcondition: if not at end of range, x points to q's next outgoing
//                transition (q,a',p') otherwise pointed transition is
//                a sink transition.
// Complexity:    constant

// Method:        find
// Semantic:      point to outoing transition labelled with specified letter
// Signature:     bool find(unsigned int b);
// Expression:    x.find('a');
// Return value:  true if transition is defined, false otherwise
// Precondition:  q != sink
// Postcondition: if transition (q,b,d1(q,b)) is defined then x points
//                to it otherwise x remains on (q,a,p)
// Complexity:    log(d2(q))

// Method:        letter
// Semantic:      access to currently pointed transition label
// Signature:     unsigned int letter() const;
// Expression:    x.letter();
// Return value:  a
// Precondition:  x must not point to a sink transition
// Postcondition: none
// Complexity:    constant

// Method:        src
// Semantic:      access to currently pointed transition source
// Signature:     State src() const;
// Expression:    x.src();
// Return value:  q
// Precondition:  none
// Postcondition: none
// Complexity:    constant

// Method:        aim
// Semantic:      access to currently pointed transition aim
// Signature:     State aim() const;
// Expression:    x.aim();
// Return value:  p
// Precondition:  x must not point to a sink transition
// Postcondition: none
// Complexity:    constant

// Method:        src_final
// Semantic:      is currently pointed transition source a final state ?
// Signature:     bool src_final() const;
// Expression:    x.src_final();
// Return value:  true if q in F, false otherwise
// Precondition:  q != sink
// Postcondition: none
// Complexity:    constant

// Method:        aim_final
// Semantic:      is currently pointed transition aim a final state ?
// Signature:     bool aim_final() const;
// Expression:    x.aim_final();
// Return value:  true if p in F, false otherwise
// Precondition:  x must not point to a sink transition
// Postcondition: none
// Complexity:    constant

// Method:        comparison
// Semantic:      is x pointing to the same transition as y ?
// Signature:     bool operator== (const forward_cursor &y) const;
// Expression:    (x == y)
// Return value:  true if (qx,ax,px)=(qy,ay,py), false otherwise
// Precondition:  none
// Postcondition: none
// Complexity:    constant



// Class stack_cursor: a stack of forward_cursor (adapter)
//  The interface manipulates the top of the stack which is a forward_cursor
//  It has exactly the same behavior as a forward cursor except that the forward method
//  pushes the new forward cursor onto the stack.
//  Method backward <=> pop
template <class ForwardCursor, template <class T> class Container = list>
class stack_cursor
{
public:
  typedef stack_cursor self;
  typedef typename DFA::Tag Tag;

 
  stack_cursor(const DFA &a, unsigned int p);     // same default behavior as forward_cursor
  stack_cursor(const DFA &a, unsigned int p, unsigned int l);     // points to transition with 'l'
  stack_cursor(const DFA &a, unsigned int p, typename DFA::Edges::const_iterator x);
  unsigned int letter() const;
  bool forward(unsigned int a);
  void forward();
  bool first_transition();
  bool next_transition();
  bool backward();                    // pop. return false if RESULTING stack is empty
  bool find(unsigned int a);
  bool src_final() const;
  bool aim_final() const;
  State src() const;
  State aim() const;
  bool operator == (const self &x) const;  // compares WHOLE stacks

  // Optional requirements:
  typedef Tag;
  self& operator= (State q);           // set the cursor on the specified state
  const Tag& src_tag() const;          // return q tag
  const Tag& aim_tag() const;          // return p tag
};  

// Class depth_first_cursor: a stack_cursor + a depth first search algorithm
//  1. A depth_first_cursor iterates on transition in a depth first order
//  2. A depth_first_cursor never pushes unusable transitions (the stack holds no unusable 
//     transition). If a depth_first_cursor reaches an unusable transition, it is not pushed
//     and the action is conceptually equivalent to a pop (forward() returns false).
//  3. A depth_first_cursor represents an algorithm stage, that is a path rather than a position in an
//     automaton. Hence the operator== behavior.

template <class DFA, template <class T> class container = list>
class depth_first_cursor : public stack_cursor<DFA, container>
{
public:
  typedef depth_first_cursor self;

  depth_first_cursor(const DFA &a);              // empty stack, end of range for algorithms
  depth_first_cursor(const DFA &a, unsigned int q);
  depth_first_cursor(const DFA &a, unsigned int q, unsigned int l);
  depth_first_cursor(const DFA &a, unsigned int q, typename DFA::Edges::const_iterator x);
  bool forward();          // move on to the NEXT STAGE OF
                           // ALGORITHM. return true if it actually
                           // moved forward (push), false if it moved backward (pop)
  unsigned int letter() const;
  bool src_final() const;
  bool aim_final() const;
  State src() const;
  State aim() const;
  bool operator == (const self &x) const;  // compares stack_cursors (compares WHOLE stacks)

  // Optional requirements:
  typedef Tag;
  const Tag& src_tag() const;          // return q tag
  const Tag& aim_tag() const;          // return p tag
};

#endif // ASTL_20_CLASS_CURSORS







