#ifndef ASTL_CURSOR_DEBUG
#define ASTL_CURSOR_DEBUG

// A default forward cursor checking its state before any processing

template <class DFA>
class forward_cursor_debug 
{
public:
  typedef forward_cursor_debug self;
  typedef typename DFA::State  State;
  typedef typename DFA::Tag    Tag;

protected:
  typedef typename DFA::Edges::const_iterator edges_iterator;
  const DFA *dfa;
  State q;
  edges_iterator transition;
  bool dereferenceable, singular;

public:
  forward_cursor_debug(const DFA &a, State p, unsigned int Letter) 
    : dfa(&a), q(p), dereferenceable(false), singular(false)
  { 
    if (q == dfa->null_state) {
      cerr 
	<< "forward_cursor::forward_cursor(DFA, State, letter): warning" << endl
	<< "\t cursor initialized with null state" 
	<< endl;
      singular = true;
    }
    else
      if ((transition = dfa->delta2(q).find(Letter)) == dfa->delta2(q).end()) {
	cerr << "forward_cursor::forward_cursor(DFA, State, letter): warning" << endl;
	cerr << "\t cursor set on a sink transition" 
	     << endl; 
      }
      else
	dereferenceable = true;
  }
  
  forward_cursor_debug(const DFA &a, State p, edges_iterator t)
    : dfa(&a), q(p), transition(t)
  { 
    if (q == dfa->null_state) {
      cerr 
	<< "forward_cursor::forward_cursor(DFA, State, edges_iterator):";
      cerr << " warning" << endl << "\t cursor initialized with null state" <<
	endl;
      singular = true;
    } 
    else 
      dereferenceable = t != dfa->delta2(q).end();
  }
  
  forward_cursor_debug(const DFA &a, State p)
    : dfa(&a), q(p)
  { 
    if (q == dfa->null_state) {
      cerr 
	<< "forward_cursor::forward_cursor(DFA, State): warning" << endl;
      cerr << "\t cursor initialized with null state" << endl;
      singular = true;
    } 
    cerr 
      << "forward_cursor::forward_cursor(DFA, State): caution" << endl;
    cerr << "\t cursor not dereferenceable" << endl; 
  }

  forward_cursor_debug(const DFA &a)
    : dfa(&a), q(a.initial())
  { 
    if (q == dfa->null_state) {
      cerr 
	<< "forward_cursor::forward_cursor(DFA): warning" << endl;
      cerr << "\t cursor initialized with initial state which is null" << endl;
      singular = true;
    } 
  }

  self& operator= (State p) {
    q = p;
    if (q == dfa->null_state) {
      cerr << "forward_cursor::operator= (State): warning" << endl;
      cerr << "\t assignment to null state" <<
	endl;
    }
    return *this;
  }

  bool operator== (const self &x) const {
    return (q == x.q && transition == x.transition);
  }

  unsigned int letter() const {
    if (sink()) {
      cerr << "forward_cursor::letter() : warning" << endl
	   << "\t trying to access sink transition of ";
      if (singular) cerr << "singular";
      else
	if (!dereferenceable) cerr << "not dereferenceable";
      cerr << " cursor pointing on state" << q << endl;
    }
    return (*transition).first;
  }

  bool sink() const {
    return q == dfa->null_state || transition == dfa->delta2(q).end();
  }

  bool first_transition() {
    singular = false;
    if (q == dfa->null_state) {
      cerr << "forward_cursor::first_transition() : warning" << endl
	   << "\t trying to access first transition of null state" << endl;
    }
    return dereferenceable = (transition = dfa->delta2(q).begin()) != dfa->delta2(q).end();
  }
  
  bool next_transition() {
    if (singular || sink()) {
      cerr << "forward_cursor::next_transition() : warning" << endl
	   << "\t trying to access sink transition of ";
      if (singular) cerr << "singular";
      else
	if (!dereferenceable) cerr << "not dereferenceable";
      cerr << " cursor pointing on state" << q << endl;
    }
    return dereferenceable = (++transition != dfa->delta2(q).end());
  }

  void forward() {
    if (sink()) {
      cerr << "forward_cursor::forward() : warning" << endl
	   << "\t trying to move along sink transition of";
      if (singular) cerr << "singular";
      else
	if (!dereferenceable) cerr << "not dereferenceable";
      cerr << " cursor pointing on state" << q << endl;
    }
    q = (*transition).second;
    dereferenceable  = false;
  }

  bool find(unsigned int letter)  {
    if (q == dfa->null_state) {
      cerr << "forward_cursor::find(a) : warning" << endl
	   << "\t trying to access outgoing transitions of sink state" << endl;
    }
    edges_iterator tmp = dfa->delta2(q).find(letter);
    if (tmp != dfa->delta2(q).end()) {
      transition = tmp;
      dereferenceable = true;
      return true;
    }
    return false;
  }

  bool forward(unsigned int letter) {
    if (q == dfa->null_state) {
      cerr << "forward_cursor::forward(a) : warning" << endl
	   << "\t trying to access outgoing transitions of sink state" << endl;
    }
    dereferenceable = false;
    if (find(letter)) {
	forward();
	return true;
      }
    return false;
  }

  State src() const {
    return q;
  }

  State aim() const {
    if (singular || !dereferenceable) {
      cerr << "forward_cursor::aim() : warning" << endl
	   << "\t trying to access outgoing transitions of";
      if (singular) cerr << "singular";
      else
	if (!dereferenceable) cerr << "not dereferenceable";
      cerr << " cursor pointing on state" << q << endl;
    }
    return (*transition).second;
  }

  bool src_final() const {
    if (q == dfa->null_state) 
      cerr << "forward_cursor::src_final() : warning" << endl
	   << "\t trying to access a sink state" << endl;
    return dfa->final(q);
  }

  bool aim_final() const {
    if (singular || !dereferenceable) {
      cerr << "forward_cursor::aim_final() : warning" << endl
	   << "\t trying to access outgoing transitions of";
      if (singular) cerr << "singular";
      else
	if (!dereferenceable) cerr << "not dereferenceable";
      cerr << " cursor pointing on state" << q << endl;
    }

    return dfa->final((*transition).second);
  }

  const Tag& src_tag() const  {
    if (q == dfa->null_state)
      cerr << "forward_cursor::src_tag() : warning" << endl
	   << "\t trying to access a sink state" << endl;
    return dfa->tag(q);
  }

  const Tag& aim_tag() const {
    if (singular || !dereferenceable) {
      cerr << "forward_cursor::aim_tag() : warning" << endl
	   << "\t trying to access outgoing transitions of";
      if (singular) cerr << "singular";
      else
	if (!dereferenceable) cerr << "not dereferenceable";
      cerr << " cursor pointing on state" << q << endl;
    }
    return dfa->tag((*transition).second);
  }

  // Not a standard requirement:
  friend ostream& operator << (ostream &out, const self &x) {
    out << x.q << " , '";
    if (x.singular) out << "singular'";
    else
      if (!x.dereferenceable) {
	if (x.transition == x.dfa->delta2(x.q).end()) out << "end of range'";
	else
	  out << "not dereferenceable'";
      }
      else
      out << (*(x.transition)).first << "', " << (*(x.transition)).second;
    return out;
  }

  // Not a standard requirement:
  bool operator < (const self &x) const {   
    // STL containers need operator < on
    // their value_type (to define < on themselves)
    if (singular) 
      cerr << "forward_cursor::operator< (forward_cursor) : caution" << endl
	   << "\t comparing singular left cursor with"
	   << (x.singular ? " singular " : " ") << "right cursor" << endl; 
    else
      if (x.singular)
	cerr << "forward_cursor::operator< (forward_cursor) : caution" << endl
	     << "\t comparing left cursor with singular right cursor" << endl; 
    return q < x.q || (q == x.q && transition < x.transition);
  }
};

#endif // ASTL_CURSOR_DEBUG
