#ifndef ASTL_CLASS_STRING_CURSORS
#define ASTL_CLASS_STRING_CURSORS

// Defines:
// template string cursor (forward cursor)

// A string_cursor simulates a flat automaton recognizing one specified word
// It is specialized for pointers and for C string (pointer+null terminated array)

// Instanciation parameters:
// 1. A forward iterator type
//
// Requirements:
// None
//
// Constructor:
// 1. A range over a word

template <class ForwardIterator>
class string_cursor
{
protected:
  ForwardIterator first, last;
  bool sink_state;

public:
  typedef ForwardIterator State;

  string_cursor(ForwardIterator first, ForwardIterator last)
    : first(first), last(last), sink_state(false)
    { }

  bool sink() const {
    return sink_state;
  }

  void forward() {
    ++first;
  }

  bool forward(unsigned int letter) {
    if (*first == letter) {
      ++first;
      return true;
    }
    return false;
  }

  bool first_transition() {
    sink_state = first == last;
    return !sink_state;
  }

  bool next_transition() {
    sink_state = true;
    return false;
  }

  unsigned int letter() const {
    return *first;
  }

  State src() const {
    return first;
  }

  State aim() const {
    ForwardIterator tmp = first;
    return ++tmp;
  }

  bool aim_final() const {
    ForwardIterator tmp = first;
    return ++tmp == last;
  }
  
  bool src_final() const {
    return first == last;
  }

  bool operator== (const string_cursor &x) const {
    return first == x.first;
  }

  bool find(unsigned int letter) {
    return *first == letter;
  }
};


// C string specialization (pointer on char + null terminated string)
// Instanciation parameters:
// 1. char*
//
// Requirements:
// None
//
// Constructor:
// 1. A C string (null terminated)

template <>
class string_cursor<char*>
{
protected:
  const char *p;
  bool sink_state;

public:
  typedef const char* State;

  string_cursor(const char *p)
    : p(p), sink_state(false)
    { }

  bool sink() const {
    return sink_state;
  }

  void forward() {
    ++p;
  }

  bool forward(unsigned int letter) {
    if ((unsigned char) *p == letter) {
      ++p;
      return true;
    }
    return false;
  }

  bool first_transition() {
    sink_state = *p == 0;
    return !sink_state;
  }

  bool next_transition() {
    sink_state = true;
    return false;
  }

  unsigned int letter() const {
    return (unsigned char) *p;   // Take care of negative value
  }

  State src() const {
    return p;
  }

  State aim() const {
    return p + 1;
  }

  bool src_final() const {
    return *p == 0;
  }

  bool aim_final() const {
    return p[1] == 0;
  }

  bool operator== (const string_cursor<char*> &x) const {
    return p == x.p;
  }

  bool find(unsigned int letter) {
    return (unsigned char) *p == letter;
  }
};

// C string specialization (pointer on unsigned char + null terminated string)
// Instanciation parameters:
// 1. unsigned char*
//
// Requirements:
// None
//
// Constructor:
// 1. A unsigned C string (null terminated)

template <>
class string_cursor<unsigned char*>
{
protected:
  const unsigned char *p;
  bool sink_state;

public:
  typedef const unsigned char* State;

  string_cursor(const unsigned char *p)
    : p(p), sink_state(false)
    { }

  void forward() {
    ++p;
  }

  bool forward(unsigned int letter) {
    if (*p == letter) {
      ++p;
      return true;
    }
    return false;
  }

  bool first_transition() {
    sink_state = *p == 0;
    return !sink_state;
  }

  bool next_transition() {
    sink_state = true;
    return false;
  }

  unsigned int letter() const {
    return *p;
  }

  State src() const {
    return p;
  }

  State aim() const {
    return p + 1;
  }

  bool src_final() const {
    return *p == 0;
  }

  bool aim_final() const {
    return p[1] == 0;
  }

  bool operator== (const string_cursor<unsigned char*> &x) const {
    return p == x.p;
  }

  bool find(unsigned int letter) {
    return *p == letter;
  }
};

#endif // ASTL_CLASS_STRING_CURSORS








